from rich.console import Console
console = Console()
from uuid import uuid4

from protocols.ldap import safe_ldap_attr, safe_ldap_sid

class Trusts:
    name = "trusts"
    desc = "Get Domain Trusts"
    search_filter = "(objectClass=trustedDomain)"
    attributes = ["cn", "distinguishedName", "securityIdentifier", "trustDirection", "trustType"]

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        results = conn.search(base_dn, self.search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        # Source: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/5026a939-44ba-47b2-99cf-386a9e674b04

        def fmt_trust_direction(direction):
            trust_direction_map = {
                0: "TRUST_DIRECTION_DISABLED [0] (-)",
                1: "TRUST_DIRECTION_INBOUND [1] (->)",
                2: "TRUST_DIRECTION_OUTBOUND [2] (<-)",
                3: "TRUST_DIRECTION_BIDIRECTIONAL [3] (<->)",
            }

            for value, description in trust_direction_map.items():
                if direction == value:
                    return str(description)
                
        def fmt_trust_type(type):
            trust_type_map = {
                1: "The trusted domain is a Windows domain not running Active Directory",
                2: "The trusted domain is a Windows domain running Active Directory",
                3: "The trusted domain is running a non-Windows, RFC4120-compliant Kerberos distribution",
                4: "Historical reference; this value is not used in Windows",
                5: "The trusted domain is in Azure Active Directory",
            }

            for value, description in trust_type_map.items():
                if type == value:
                    return str(description)

        values = []
        for entry in conn.entries:
            cn = safe_ldap_attr(entry, 'cn', 'None')
            distinguishedName = safe_ldap_attr(entry, 'distinguishedName', 'None')
            securityIdentifier = safe_ldap_sid(entry, 'securityIdentifier', 'None')

            raw_direction = safe_ldap_attr(entry, 'trustDirection', None)
            if raw_direction is not None:
                trustDirection = fmt_trust_direction(raw_direction)
            else:
                trustDirection = 'None'

            raw_type = safe_ldap_attr(entry, 'trustType', None)
            if raw_type is not None:
                trustType = fmt_trust_type(raw_type)
            else:
                trustType = 'None'

            result = f"""cn: {cn}
distinguishedName: {distinguishedName}
SID: {securityIdentifier}
Trust Direction: {trustDirection}
Trust Type: {trustType}\n
"""

            values.append(result)
            console.print(result, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from rich.console import Console
console = Console()
from uuid import uuid4
from ldap3.core.exceptions import LDAPInsufficientAccessRightsResult

class Changepass:
    name = "changepass"
    desc = "Change desired user password"
    attributes = ["sAMAccountName"]

    def get_user_dn(self, conn, base_dn, user):
        conn.search(base_dn, f"(&(objectClass=user)(sAMAccountName={user}))", attributes=["distinguishedName"])
        for entry in conn.entries:
            user_dn = entry.distinguishedName.value or "None"
            return str(user_dn)

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        if not module_args or len(module_args) < 2:
            console.print("\n[[red]![/]] You need to specify a username and a password. Example: hexodus <profile> ldap <host> user [yellow]Administrator[/] [yellow]NewPass123[/]", highlight=False)
            return

        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        username = module_args[0]
        password = module_args[1]
        user_dn = self.get_user_dn(conn, base_dn, username)
        
        if user_dn == "None":
            console.print(f"\n[[red]x[/]] Impossible to get {username} distinguishedName attribute value. Check if the user actually exists.")
            return
        
        console.print(f"\n[yellow]{username} distinguishedName: {user_dn}[/]", highlight=False)
        console.print(f"[[yellow]![/]] Changing {username} password to {password}", highlight=False)

        try:
            change_pass = conn.extend.microsoft.modify_password(user=user_dn, new_password=password, old_password=None)

            if not(change_pass):
                console.print(f"[[red]x[/]] Unable to change {username} password to {password}", highlight=False)
                return

        except LDAPInsufficientAccessRightsResult:
            console.print(f"[[red]x[/]] You don't have enough privileges to change [b]{username}[/] password", highlight=False)
            return

        values = []
        result = f"{username}'s password changed to {password} successfully"
        console.print(result, highlight=False)
        values.append(result)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from rich.console import Console
console = Console()
from uuid import uuid4
from ldap3 import MODIFY_ADD
from ldap3.core.exceptions import LDAPInsufficientAccessRightsResult, LDAPEntryAlreadyExistsResult

from protocols.ldap import safe_ldap_attr

class Add_to_group:
    name = "addtogroup"
    desc = "Add user to a desired group"

    def get_user_dn(self, conn, base_dn, user):
        conn.search(base_dn, f"(&(objectClass=user)(sAMAccountName={user}))", attributes=["distinguishedName"])
        for entry in conn.entries:
            user_dn = safe_ldap_attr(entry, 'distinguishedName', 'None')
            return str(user_dn)

    def get_group_dn(self, conn, base_dn, group):
        conn.search(base_dn, f"(&(objectClass=group)(cn={group}))", attributes=["distinguishedName"])
        for entry in conn.entries:
            group_dn = safe_ldap_attr(entry, 'distinguishedName', 'None')
            return str(group_dn)

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        if not module_args or len(module_args) < 2:
            console.print("\n[[red]![/]] You need to specify a username and a grouop. Example: hexodus <profile> ldap <host> user [yellow]Administrator[/] [yellow]'Domain Admins'[/]", highlight=False)
            return

        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        username = module_args[0]
        group = module_args[1]

        user_dn = self.get_user_dn(conn, base_dn, username)
        group_dn = self.get_group_dn(conn, base_dn, group)

        console.print(f"\n[[yellow]![/]] Adding {username} to {group} group", highlight=False)

        try:
            add_to_group = conn.modify(group_dn, {"member": [(MODIFY_ADD, [user_dn])]})

        except LDAPEntryAlreadyExistsResult:
            console.print(f"[[yellow]![/]] {username} is already a member of {group} group", highlight=False)
            return

        except LDAPInsufficientAccessRightsResult:
            console.print(f"[[red]x[/]] You don't have enough privileges to add {username} to {group} group", highlight=False)
            return


        values = []
        result = f"{username} added to {group} successfully"
        console.print(result, highlight=False)
        values.append(result)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from rich.console import Console
console = Console()
from uuid import uuid4
from impacket.dcerpc.v5 import srvs

from protocols.rpc import get_rpc_connection

class Share:
    name = "share"
    desc = "Enumerates the available shares of a target computer"

    def on_login(self, host, save_output = False, module_args = None):

        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module\n")

        rpc_connection = get_rpc_connection(host)
        resp = srvs.hNetrShareEnum(rpc_connection, 2)

        if rpc_connection is None:
                console.print(f"[[red]![/]] Unable to estabilish RPC connection with {host}")
                return

        if resp is None or resp["ErrorCode"] != 0:
            console.print("[[red]![/]] Error enumerating share information")
            return
        
        output = []

        for share in resp["InfoStruct"]["ShareInfo"]["Level2"]["Buffer"]:
            share_name = share["shi2_netname"][:-1]
            share_remark = share["shi2_remark"][:-1]
            share_path = share["shi2_path"][:-1]

            share_info: str = f"{share_name} - {share_remark} - {share_path}"
            
            console.print(f"{share_info} ", highlight=False)

            output.append(share_info)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as file:
                for value in output:
                    file.write(value + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return output



import ldap3
import ssl
from ldap3.core.exceptions import LDAPBindError, LDAPInvalidCredentialsResult, LDAPCursorAttributeError
from rich.console import Console
from sys import exit

from handlers.profile import get_username, get_password, get_domain
from parsers.formatters import fmt_sid

console = Console()

def get_ldap_connection(host: str):
    """ LDAP Connection Handler: tries signing on 389, then LDAPS on 636. """

    username, password, domain = get_username(), get_password(), get_domain()

    user = f"{domain}\\{username}"

    if len(password) == 32 and all(c in "0123456789abcdefABCDEF" for c in password):
        password = f"aad3b435b51404eeaad3b435b51404ee:{password}"

    tls = ldap3.Tls(validate=ssl.CERT_NONE,
          version=ssl.PROTOCOL_TLSv1_2,
          ciphers="ALL:@SECLEVEL=0")
    
    ldaps_server = ldap3.Server(f"ldaps://{host}", port=636,
                    use_ssl=True, get_info=ldap3.ALL, tls=tls)

    ldap_server = ldap3.Server(f"ldap://{host}", port=389,
                    use_ssl=False, get_info=ldap3.ALL)

    try:
        ldap_connection = ldap3.Connection(
            server=ldap_server,
            user=user,
            password=password,
            authentication=ldap3.NTLM,
            auto_bind=True,
            session_security="ENCRYPT", # ← enable LDAP signing
            auto_referrals=False,
            raise_exceptions=True
        )

        signing = "[green]Yes[/]" if ldap_connection.session_security in ("SIGNATURE", "ENCRYPT") else "[red]No[/]"
        encrypted = "[green]Yes[/]" if ldap_connection.server.ssl else "[red]No[/]"
        #base_dn = ldap_connection.server.info.other.get("defaultNamingContext", ["<none>"])[0]
        base_dn = ldap_connection.server.info.naming_contexts[0]
        
        console.print(f"[[green]+[/]] [cyan]LDAP[/]    {host} " f"({base_dn}) " f"(Signing: {signing}) " f"(Encrypted: {encrypted})", highlight=False)
        console.print(f"[[green]+[/]] [cyan]LDAP[/]    {domain}\\{username}:{password}", highlight=False)
        console.print("[[green]+[/]] [cyan]LDAP[/]    bind successful with signing", highlight=False)

        return ldap_connection, base_dn

    except LDAPBindError as e:
        if "strongerAuthRequired" not in str(e):
            console.print(f"[[red]x[/]] LDAP bind error: {e}", highlight=False)
            raise

    except LDAPInvalidCredentialsResult:
        console.print(f"[[red]x[/]] Invalid credentials provided. ({user}:{password})", highlight=False)
        exit(0)

    try:
        ldaps_connection = ldap3.Connection(
            server=ldaps_server,
            user=user,
            password=password,
            authentication=ldap3.NTLM,
            auto_bind=True,
            auto_referrals=False,
            raise_exceptions=True
        )

        signing = "[green]Yes[/]" if ldaps_connection.session_security in ("SIGNATURE", "ENCRYPT") else "[red]No[/]"
        encrypted = "[green]Yes[/]" if ldaps_connection.server.ssl else "[red]No[/]"
        base_dn = ldaps_connection.server.info.other.get("defaultNamingContext", ["<none>"])[0]

        console.print(f"[[green]+[/]] [cyan]LDAPS[/]    {host} " f"({base_dn}) " f"(Signing: {signing}) " f"(Encrypted: {encrypted})", highlight=False)
        console.print(f"[[green]+[/]] [cyan]LDAPS[/]    {domain}\\{username}:{password}", highlight=False)
        console.print("[[green]+[/]] [cyan]LDAPS[/]     bind successful", highlight=False)

        return ldaps_connection, base_dn
    
    except LDAPInvalidCredentialsResult:
        console.print(f"[[red]x[/]] Invalid credentials provided. ({user}:{password})", highlight=False)
        exit(0)

    except LDAPBindError as e:
        console.print(f"[[red]x[/]] LDAPS bind failed: {e}", highlight=False)
        raise


def safe_ldap_attr(entry, attr_name, fallback=None) -> None:
    """ Safely get a LDAP attribute value or return a valid fallback to avoid exceptions """
    try:
        attr = getattr(entry, attr_name, None)
        return attr.value if attr else fallback
    except (AttributeError, LDAPCursorAttributeError):
        return fallback

def safe_ldap_sid(entry, attr_name='securityIdentifier', default='None'):
    """ Safely get and format SID from LDAP entry """
    try:
        raw_sid = safe_ldap_attr(entry, attr_name, None)
        if raw_sid:
            return fmt_sid(raw_sid)
        return default
    except Exception as e:
        print(f"Error formatting SID: {e}")
        return default


from importlib import resources, import_module
from rich.console import Console
console = Console()
from argparse import ArgumentParser, REMAINDER
from os import environ, path

from handlers.environment import *
from protocols.smb import get_smb_connection
from protocols.ldap import get_ldap_connection
from webapp import start_flask, stop_flask

from sys import argv

def setup() -> None:
    """
    Initialize and manage the Hexodus environment and CLI operations.
    """

    parser = ArgumentParser(
        description="To list modules from a protocol use: hexodus <profile> <protocol> list"
    )

    if len(argv) == 1:
        parser.print_help()
        return

    # Positional argument to be used as the PROFILE
    parser.add_argument(
        "profile",
        help="Specify the profile to be used",
        nargs="?", 
        default=None
    )

    # Positional argument to be used as the PROTOCOL
    parser.add_argument(
        "protocol",
        help="Specify the protocol to be used",
        nargs="?", 
        default=None
    )

    # Positional argument to be used as the HOST
    parser.add_argument(
        "host",
        help="Specify the host to be used",
        nargs="?",
        default=None
    )

    # Positional argument to be used as the MODULE
    parser.add_argument(
        "module",
        help="Specify the module to be used",
        nargs="?", 
        default=None
    )

    parser.add_argument(
        "module_args",
        help="Arguments specific to the chosen module",
        nargs='*',
        default=None
    )

    # Create profile
    parser.add_argument(
        "--create-profile", "-c",
        metavar=('PROFILE_NAME', 'PROFILE_DOMAIN', 'PROFILE_USERNAME', 'PROFILE_PASSWORD'),
        help="Usage: hexodus --create-profile <profile_name> <profile_domain> <profile_username> <profile_password>", 
        nargs=4, 
        default=[]
    )
    
    # Delete profile
    parser.add_argument(
        "--delete-profile", "-d",
        metavar=('PROFILE_NAME'),
        help="Usage: hexodus --delete-profile <profile_name>", 
    )

    # List profiles
    parser.add_argument(
        "--list-profiles", "-lp",
        help="Usage: hexodus --list-profiles",
        action="store_true"
    )

    # Save the output of the command
    parser.add_argument(
        "--output", "-o",
        help="Usage: hexodus <profile> <protocol> <host> <module> --output",
        action="store_true"
    )

    # Generate a web view of the collected data
    parser.add_argument(
        "--start-web", "-s",
        help="Usage: hexodus --start-web",
        action="store_true"
    )

    # Stop the Flask web app
    parser.add_argument(
        "--stop-web", "-sw",
        help="Usage: hexodus --stop-web",
        action="store_true"
    )

    # Delete the profile column from the data.db file
    parser.add_argument(
        "--delete-profile-column", "-x",
        metavar="PROFILE_NAME",
        help="Usage: hexodus --delete-profile-column <profile-name>",
        type=str,
    )

    args = parser.parse_args()
    save_output = args.output
    ensure_base_dir()
    
    if args.create_profile:
        create_profile(*args.create_profile)

    if args.delete_profile:
        delete_profile(args.delete_profile)

    if args.list_profiles:
        list_profiles()
        return
    
    if args.start_web:
        start_flask()
        return

    if args.stop_web:
        stop_flask()
        return
    
    if args.delete_profile_column:
        drop_profile_column(args.delete_profile_column)
        return

    if args.profile:
        # Check if profile exists before loading it
        profile_path: str = f"{user_home()}/.hexodus/{args.profile}"
        if not path.exists(profile_path):
            console.print(f"[[red]x[/]] {args.profile} does not exists.", highlight=False)
            return
        
        environ['hexodus_profile'] = args.profile
        init_database()
        ensure_profile_column(args.profile)

        if args.profile and args.protocol is None:
            console.print(f"[[green]+[/]] Loaded profile: {environ['hexodus_profile']}")
            return

        if args.protocol == "smb":

            if args.module == None and args.host == None:
                console.print("[[red]![/]] You need to specify at least one host address")
                return

            if args.module == None and args.host != "list":
                get_smb_connection(args.host)
                return
            
            if args.host == "list":
                import modules.smb as smb_pkg
                for file in resources.files(smb_pkg).iterdir():
                    if file.suffix == ".py" and file.name != "__init__.py":
                        console.print(f"[[green]+[/]] {file.stem}")

            if args.module:
                get_smb_connection(args.host)

                try:
                    module = import_module(f"modules.smb.{args.module}")
                    cls = getattr(module, args.module.capitalize())
                    instance = cls()
                    values = instance.on_login(args.host, save_output, args.module_args)

                    if values:
                        text = "\n".join(values)
                        save_to_database(args.profile, args.module, text)

                except ModuleNotFoundError:
                    raise Exception
                    console.print(f"\n[[red]x[/]] Module '{args.module}' not found")
                    return


        if args.protocol == "ldap":

            if args.module == None and args.host == None:
                console.print("[[red]x[/]] You need to specify at least one host address")
                return

            if args.module == None and args.host != "list":
                get_ldap_connection(args.host)
                return
            
            if args.host == "list":
                import modules.ldap as ldap_pkg
                for file in resources.files(ldap_pkg).iterdir():
                    if file.suffix == ".py" and file.name != "__init__.py":
                        console.print(f"[[green]+[/]] {file.stem}")

            if args.module:
                conn, base_dn = get_ldap_connection(args.host)

                try:
                    module = import_module(f"modules.ldap.{args.module}")
                    cls = getattr(module, args.module.capitalize())
                    instance = cls()
                    values = instance.on_login(conn, base_dn, save_output, args.module_args)

                    if values:
                        text = "\n".join(values)
                        save_to_database(args.profile, args.module, text)

                except ModuleNotFoundError:
                    raise Exception
                    console.print(f"\n[[red]x[/]] Module '{args.module}' not found")
                    return


from rich.console import Console
console = Console()
from uuid import uuid4

class Users:
    name = "users"
    desc = "Get 'sAMAccountName' attribute value from Users Accounts"
    search_filter = "(&(objectCategory=person)(objectClass=user))"
    attributes = "sAMAccountName"

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        try:
            generator = conn.extend.standard.paged_search(
                search_base=base_dn,
                search_filter=self.search_filter,
                search_scope='SUBTREE',
                attributes=[self.attributes],
                paged_size=1000,
                generator=True
            )

            values = []
            for entry in generator:
                if entry.get('type') != 'searchResEntry':
                    continue

                username_attribute = entry.get('attributes', {}).get(self.attributes)
                if not username_attribute:
                    continue

                value = username_attribute[0] if isinstance(username_attribute, list) else username_attribute
                values.append(value)
                console.print(value, highlight=False)
            
            if not values:
                console.print(f"\n[[red]![/]] No results for {self.name}") 
                return
            
            if save_output:
                filename = f"{self.name}_{uuid4().hex}.txt"
                with open(filename, "w", encoding="utf-8") as f:
                    for v in values:
                        f.write(v + "\n")
                console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

            return values
        
        except Exception as error:
            console.print(f"[[red]![/]] Error in users.py module: {error}", highlight=False)
            raise


from rich.console import Console
console = Console()
from uuid import uuid4

from protocols.ldap import safe_ldap_attr

class Passpol:
    name = "passpol"
    desc = "Get the domain password policy"
    search_filter = "(objectClass=domainDNS)"
    attributes = ["lockoutDuration", "lockoutThreshold", "maxPwdAge", "minPwdAge", "minPwdLength"]

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        results = conn.search(base_dn, self.search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        values = []
        for entry in conn.entries:
            lockoutDuration = safe_ldap_attr(entry, 'lockoutDuration', 'None')
            lockoutThreshold = f"{safe_ldap_attr(entry, 'lockoutThreshold', 'None')} - Password Spray is possible!" if safe_ldap_attr(entry, 'lockoutThreshold', 'None') == 0 else safe_ldap_attr(entry, 'lockoutThreshold', 'None') or "None"
            maxPwdAge = safe_ldap_attr(entry, 'maxPwdAge', 'None')
            minPwdAge = safe_ldap_attr(entry, 'minPwdAge', 'None')
            minPwdLength = safe_ldap_attr(entry, 'minPwdLength', 'None')

            result = f"""Lockout Duration: {lockoutDuration}
Lockout Threshold: {lockoutThreshold}
Max Password Age: {maxPwdAge}
Min Password Age: {minPwdAge}
Min Password Length: {minPwdLength}"""
            values.append(result)

            console.print(result, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as file:
                for value in values:
                    file.write(value)

            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from os import environ
from json import load
from rich.console import Console
console = Console()

from handlers.environment import user_home

def get_username() -> str:
    """ Read the 'username' property from the current profile file """

    profile = environ.get('hexodus_profile')
    profile_path: str = f"{user_home()}/.hexodus/{profile}/profile.json"

    try:
        with open(profile_path, 'r', encoding='utf-8') as json_file:
            data = load(json_file)
            return data['username']

    except Exception as error:
        console.print(f"[[red]x[/]] Error when running get_username function in profile.py: {error}")

def get_password() -> str:
    """ Read the 'password' property from the current profile file """

    profile = environ.get('hexodus_profile')
    profile_path: str = f"{user_home()}/.hexodus/{profile}/profile.json"

    try:
        with open(profile_path, 'r', encoding='utf-8') as json_file:
            data = load(json_file)
            return data['password']

    except Exception as error:
        console.print(f"[[red]x[/]] Error when running get_password function in profile.py: {error}")

def get_domain() -> str:
    """ Read the 'domain' property from the current profile file """

    profile = environ.get('hexodus_profile')
    profile_path: str = f"{user_home()}/.hexodus/{profile}/profile.json"

    try:
        with open(profile_path, 'r', encoding='utf-8') as json_file:
            data = load(json_file)
            return data['domain']

    except Exception as error:
        console.print(f"[[red]x[/]] Error when running get_domain function in profile.py: {error}")


from impacket.structure import Structure

class MSDS_MANAGEDPASSWORD_BLOB(Structure):
    structure = (
        ('Version','<H'),
        ('Reserved','<H'),
        ('Length','<L'),
        ('CurrentPasswordOffset','<H'),
        ('PreviousPasswordOffset','<H'),
        ('QueryPasswordIntervalOffset','<H'),
        ('UnchangedPasswordIntervalOffset','<H'),
        ('CurrentPassword',':'),
        ('PreviousPassword',':'),
        #('AlignmentPadding',':'),
        ('QueryPasswordInterval',':'),
        ('UnchangedPasswordInterval',':'),
    )

    def __init__(self, data = None):
        Structure.__init__(self, data = data)

    def fromString(self, data):
        Structure.fromString(self,data)

        if self['PreviousPasswordOffset'] == 0:
            endData = self['QueryPasswordIntervalOffset']
        else:
            endData = self['PreviousPasswordOffset']

        self['CurrentPassword'] = self.rawData[self['CurrentPasswordOffset']:][:endData - self['CurrentPasswordOffset']]
        if self['PreviousPasswordOffset'] != 0:
            self['PreviousPassword'] = self.rawData[self['PreviousPasswordOffset']:][:self['QueryPasswordIntervalOffset']-self['PreviousPasswordOffset']]

        self['QueryPasswordInterval'] = self.rawData[self['QueryPasswordIntervalOffset']:][:self['UnchangedPasswordIntervalOffset']-self['QueryPasswordIntervalOffset']]
        self['UnchangedPasswordInterval'] = self.rawData[self['UnchangedPasswordIntervalOffset']:]


from rich.console import Console
console = Console()
from uuid import uuid4

from parsers.formatters import fmt_gmsa
from protocols.ldap import safe_ldap_attr

class Gmsa:
    name = "gmsa"
    desc = "Get GMSA accounts passwords"
    search_filter = "(objectClass=msDS-GroupManagedServiceAccount)"
    attributes = ["sAMAccountName", "msDS-ManagedPassword"]

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        results = conn.search(base_dn, self.search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        values = []
        for entry in conn.entries:
            sAMAccountName = safe_ldap_attr(entry, 'sAMAccountName', 'None')
            msDS_ManagedPassword = safe_ldap_attr(entry, fmt_gmsa('msDS-ManagedPassword'), 'None')

            result = f"Account: {sAMAccountName} - NT Hash: {msDS_ManagedPassword}"
            values.append(result)

            console.print(result, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from rich.console import Console
console = Console()
from uuid import uuid4

class Nopreauth:
    name = "nopreauth"
    desc = "List all users that do not require Kerberos pre-authentication"
    search_filter = "(&(objectclass=user)(objectcategory=user)(useraccountcontrol:1.2.840.113556.1.4.803:=4194304))"
    attributes = "sAMAccountName"

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        results = conn.search(base_dn, self.search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        values = [entry[self.attributes].value for entry in conn.entries]
        for value in values:
            console.print(value, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from rich.console import Console
console = Console()
from uuid import uuid4
from os import makedirs, path, environ
from shutil import rmtree
from pathlib import Path
from json import dump

import sqlite3

def init_database() -> None:
    """ Initializes the default Hexodus database """
    
    database = path.join(user_home(), ".hexodus", "data.db")
    sqlite_cursor = sqlite3.connect(database)
    sqlite_cursor.execute("""
      CREATE TABLE IF NOT EXISTS data (
        module TEXT PRIMARY KEY
      );
    """)
    sqlite_cursor.commit()
    sqlite_cursor.close()


def ensure_profile_column(profile: str) -> None:
    """ Makes sure that the "profile" column exists in Hexodus database """
    
    database = path.join(user_home(), ".hexodus", "data.db")
    database_connection = sqlite3.connect(database)
    sqlite_cursor = database_connection.cursor()

    sqlite_cursor.execute("PRAGMA table_info(data);")
    existing = {row[1] for row in sqlite_cursor.fetchall()}

    if "module" not in existing:
        sqlite_cursor.execute("ALTER TABLE data ADD COLUMN module TEXT;")
        sqlite_cursor.execute("CREATE UNIQUE INDEX IF NOT EXISTS idx_data_module ON data(module);")

    if profile not in existing:
        sqlite_cursor.execute(f"ALTER TABLE data ADD COLUMN '{profile}' TEXT;")

    database_connection.commit()
    database_connection.close()


def save_to_database(profile: str, module: str, output: str) -> None:
    """ Save the collected data to the database to be used by the webapp """

    database = f"{user_home()}/.hexodus/data.db"
    database_connection = sqlite3.connect(database)
    sqlite_cursor = database_connection.cursor()

    sqlite_cursor.execute("""
      INSERT INTO data(module) VALUES(?)
      ON CONFLICT(module) DO NOTHING
    """, (module,))

    sqlite_cursor.execute(f"""
      UPDATE data
      SET "{profile}" = ?
      WHERE module = ?
    """, (output, module))

    database_connection.commit()
    database_connection.close()


def drop_profile_column(profile_name: str) -> None:
    """ Remove the specified profile data from the database """

    database = f"{user_home()}/.hexodus/data.db"
    database_connection = sqlite3.connect(database)
    sqlite_cursor = database_connection.cursor()

    sqlite_cursor.execute(f'ALTER TABLE data DROP COLUMN "{profile_name}"')
    database_connection.commit()
    database_connection.close()


def environment_json(profile_name: str, profile_domain: str, profile_username: str, profile_password: str) -> None:
    """ Create the profile data JSON file to be used by MySQL, Django services and other functions """

    profile_name = profile_name
    profile_path: str = f"{user_home()}/.hexodus/{profile_name}"
    profile_uuid = uuid4().hex

    profile_data = {
        "name": profile_name,
        "uuid": profile_uuid,
        "username": profile_username,
        "password": profile_password,
        "domain": profile_domain,
    }

    try:
        with open(f"{profile_path}/profile.json", "w", encoding="utf-8") as json_file:
            dump(profile_data, json_file, ensure_ascii=False, indent=4)

    except Exception as error:
        console.print(f"[[red][x][/]] Error when creating the profile JSON: {error}")
        return False


def list_profiles() -> None:
    """ List existent profiles based in hexodus directory """

    profile_path: str = f"{user_home()}/.hexodus/"
    folders = [folder for folder in Path(profile_path).iterdir() if folder.is_dir()]

    if not folders:
        console.print(f"[[red]x[/]] No profiles found in {profile_path}", highlight=False)
        return False
    
    for folder in folders:
        console.print(f"* {folder}", highlight=False)
    

def delete_profile(profile_name: str) -> None:
    """ Delete target profile directory """

    profile_path: str = f"{user_home()}/.hexodus/{profile_name}"

    if not path.exists(profile_path):
        console.print(f"[[red]x[/]] {profile_path} does not exists.", highlight=False)
        return False
    
    rmtree(profile_path)

    drop_profile_column(profile_name)

    console.print(f"[[green]![/]] {profile_path} deleted.", highlight=False)
    

def create_profile(profile_name: str, profile_domain: str, profile_username: str, profile_password: str) -> None:
    """ Receive 5 string arguments to create a new profile and an JSON file """

    profile_path: str = f"{user_home()}/.hexodus/{profile_name}"
    if not path.exists(profile_path):
        makedirs(profile_path)
        console.print(f"[[green]![/]] {profile_path} created.", highlight=False)
        environment_json(profile_name, profile_domain, profile_username, profile_password)
    else:
        console.print(f"[[red]x[/]] {profile_path} already exists, use another name for the profile.", highlight=False)
        return False


def user_home() -> str:
    """ Return user's home directory as string """

    return str(Path.home())


def ensure_base_dir() -> None:
    """ Ensures hexodus directory exists in users's home """

    base_path = f"{user_home()}/.hexodus"
    if not path.exists(base_path):
        makedirs(base_path)

    if not path.exists(f"{base_path}/data.db"):
        init_database()


from struct import unpack_from
from Cryptodome.Hash import MD4
from binascii import hexlify

from parsers.gmsa import MSDS_MANAGEDPASSWORD_BLOB

def fmt_multi(input):
    """ Format values that can be str, list or None and returns an formatted string """

    if isinstance(input, list) and input:
        return ", ".join(input)
    
    if isinstance(input, str) and input:
        return input
    
    return "None"

def fmt_uac(value):
    """ Get UAC value and format to understandable UAC status """

    uac_values = {
        "512": "User is Enabled - Password Expires",
        "514": "User is Disabled - Password Expires",
        "66048": "User is Enabled - Password Never Expires",
        "66050": "User is Disabled - Password Never Expires",
        "1114624": "User is Enabled - Password Never Expires - User Not Delegated",
        "1049088": "User is Enabled - Password Expires - User Not Delegated",
        "17891840": "User is Enabled - [bold yellow]Password Never Expires - [bold yellow]User Trusted to Delegate",
        "66176": "Password Never Expires - ms-DS-User-Encrypted-Text-Password-Allowed is true",
        "4096": "WORKSTATION_TRUST_ACCOUNT",
    }

    for uac_value, description in uac_values.items():
        if int(uac_value) == value:
            return description
        
def fmt_sid(sid):
    """ Convert the SID binary to a readable string """

    revision, sub_authority_count = unpack_from("BB", sid, 0)
    identifier_authority = unpack_from(">Q", b"\x00\x00" + sid[2:8])[0]
    sub_authorities = unpack_from(f"<{sub_authority_count}I", sid, 8)

    sid_string = f"S-{revision}-{identifier_authority}"
    for sub_authority in sub_authorities:
        sid_string += f"-{sub_authority}"

    return sid_string

def fmt_gmsa(password):
    """ Transform GMSA blob password to usable NT hash """

    blob = MSDS_MANAGEDPASSWORD_BLOB()

    if len(password) != 0:
        blob.fromString(password)
        hash = MD4.new()
        hash.update(blob['CurrentPassword'][:-2])
        msDS_ManagedPassword_cleartext = hexlify(hash.digest()).decode("ascii")
        return msDS_ManagedPassword_cleartext
    
    return "None"


from rich.console import Console
console = Console()
from uuid import uuid4

from protocols.ldap import safe_ldap_attr

class Search_computer:
    name = "search_computer"
    desc = "Get information from the group name specified"
    attributes = ["dNSHostName"]

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        if not module_args or len(module_args) == 0:
            console.print("\n[[red]![/]] You need to specify a group name. Example: hexodus <profile> ldap <host> searchcomputer [yellow]'SRVWEB'[/]", highlight=False)
            return

        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        computer_name = " ".join(module_args)
        search_filter = f"(&(objectCategory=computer)(cn=*{computer_name}*))"
        results = conn.search(base_dn, search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [black]{search_filter}[/]\n", highlight=False)

        values = []
        for entry in conn.entries:
            dNSHostName = safe_ldap_attr(entry, 'dNSHostName', 'Blank')
            
            values.append(dNSHostName)
            console.print(dNSHostName, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from impacket.smbconnection import SMBConnection, SessionError
from rich.console import Console
console = Console()

from handlers.profile import get_username, get_password, get_domain

def get_smb_connection(host: str) -> None:
    """ SMB Connection Handler: Establish an SMB connection with the host and return the connection object """

    username, password, domain = get_username(), get_password(), get_domain()

    try:
        smb_connection = SMBConnection(host, host)

        if len(password) == 32 and all(c in "0123456789abcdefABCDEF" for c in password):
            lmhash = "aad3b435b51404eeaad3b435b51404ee"
            nthash = password
            smb_connection.login(username, '', lmhash=lmhash, nthash=nthash)
        else:
            smb_connection.login(username, password, domain)

        smb_name = smb_connection.getServerDNSHostName()
        smb_os = smb_connection.getServerOS()
        smb_ntlmv2_support: bool = '[green]True[/]' if smb_connection.doesSupportNTLMv2() else '[red]False[/]'
        smb_signing_required: bool = '[green]True[/]' if smb_connection.isSigningRequired() else '[red]False[/]'

        console.print(f"[[green]+[/]] [cyan]SMB[/]     {host} {smb_os} (Name: {smb_name}) (Signing: {smb_signing_required}) (NTLMv2: {smb_ntlmv2_support})", highlight=False)
        console.print(f"[[green]+[/]] [cyan]SMB[/]     {domain}\\{username}:{password}", highlight=False)

        return smb_connection
    
    except SessionError as session_error:
        if "STATUS_LOGON_FAILURE" in str(session_error):
            console.print(f"[[red]x[/]] Invalid credentials: [cyan]SMB[/]     {host} {domain}\\{username}:{password}", highlight=False)
        return None

    except Exception as error:
        console.print(f"[[red]x[/]] Error establishing SMB connection: {error}", highlight=False)
        return None   


import sqlite3
import argparse
import subprocess, sys, os

from flask import Flask, render_template
from rich.console import Console
console = Console()
from pathlib import Path
from signal import SIGTERM

def start_flask():
    pid_file = Path(os.path.expanduser("~")) / ".hexodus" / "flask.pid"
    pid_file.parent.mkdir(exist_ok=True)

    if pid_file.exists():
        pid = int(pid_file.read_text())
        try:
            os.kill(pid, 0)
            console.print(f"[[yellow]![/]] Web app already running on http://127.0.0.1:1337 - (PID: [yellow]{pid}[/])")
            return
        except ProcessLookupError:
            pid_file.unlink()

    proc = subprocess.Popen(
        [sys.executable, "-m", "webapp", "--port", "1337"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        start_new_session=True
    )

    pid_file.write_text(str(proc.pid))
    console.print(f"[[green]+[/]] Web app started - http://127.0.0.1:1337 - (PID: [yellow]{proc.pid}[/])")


def stop_flask():
    pid_file = Path(os.path.expanduser("~")) / ".hexodus" / "flask.pid"
    if not pid_file.exists():
        console.print("[[yellow]![/]] No running web app found")
        return

    pid = int(pid_file.read_text())
    try:
        os.killpg(pid, SIGTERM)
        console.print(f"[[green]+[/]] Web app stopped successfully ([yellow]{pid}[/])")
    except ProcessLookupError:
        console.print(f"[[red]x[/]] Web app not found ([yellow]{pid}[/])")
    finally:
        pid_file.unlink()

def create_app():
    app = Flask(__name__, template_folder=os.path.join(os.path.dirname(__file__), "templates"))

    @app.route("/")
    def index():
        db_path = os.path.join(os.path.expanduser("~"), ".hexodus", "data.db")
        conn = sqlite3.connect(db_path)
        sqlite_cursor = conn.cursor()

        sqlite_cursor.execute("PRAGMA table_info(data);")
        cols = [row[1] for row in sqlite_cursor.fetchall()]
        profiles = [c for c in cols if c not in ("module", "id")]

        sqlite_cursor.execute("SELECT * FROM data;")
        rows = sqlite_cursor.fetchall()
        conn.close()

        modules = []
        for row in rows:
            entry = { cols[i]: row[i] for i in range(len(cols)) }
            modules.append(entry)

        return render_template("index.html", modules=modules, profiles=profiles)

    return app

if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument("--port", type=int, default=1337)
    args = p.parse_args()
    app = create_app()
    app.run(host="127.0.0.1", port=args.port, debug=False)


from rich.console import Console
console = Console()
from uuid import uuid4
from ldap3 import SUBTREE

from protocols.ldap import safe_ldap_attr

class Domainsid:
    name = "domainsid"
    desc = "Get SID value from Domain Controllers"
    search_filter = "(&(objectClass=computer)(userAccountControl:1.2.840.113556.1.4.803:=8192))"
    attributes = ["cn", "objectSid", "dNSHostName"]

    def search_with_base(self, conn, search_base, search_filter, attributes, scope):
        return conn.search(
            search_base=search_base,
            search_filter=search_filter,
            search_scope=scope,
            attributes=attributes,
        )

    def on_login(self, conn, base_dn, save_output=False, module_args=None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        results = self.search_with_base(
            conn,
            search_base=base_dn,
            search_filter=self.search_filter,
            attributes=self.attributes,
            scope=SUBTREE,
        )

        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        values = []
        for entry in entries:
            dNSHostName = safe_ldap_attr(entry, 'dNSHostName', 'None')
            objectSid = safe_ldap_attr(entry, 'objectSid', 'None')
            cn = safe_ldap_attr(entry, 'cn', 'None')

            result = f"{dNSHostName} - {objectSid} - {cn}"
            console.print(result, highlight=False)

            values.append(result)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from rich.console import Console
console = Console()
from uuid import uuid4

from parsers.formatters import fmt_multi, fmt_uac
from protocols.ldap import safe_ldap_attr

class User:
    name = "user"
    desc = "Get information from the user specified"
    attributes = [
        "description",
        "memberOf",
        "userAccountControl",
        "badPwdCount",
        "lastLogoff",
        "lastLogon",
        "objectSid",
        "adminCount",
        "accountExpires",
        "sAMAccountName",
        "servicePrincipalName"
    ]

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        if not module_args or len(module_args) == 0:
            console.print("\n[[red]![/]] You need to specify an username. Example: hexodus <profile> ldap <host> user [yellow]Administrator[/]", highlight=False)
            return

        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        username = " ".join(module_args)
        search_filter = f"(&(objectClass=user)(sAMAccountName={username}))"
        results = conn.search(base_dn, search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [black]{search_filter}[/]\n", highlight=False)

        values = []
        for entry in conn.entries:
            description = safe_ldap_attr(entry, 'description', 'None')
            memberOf = safe_ldap_attr(entry, fmt_multi('memberOf', 'None'))
            objectSid = safe_ldap_attr(entry, 'objectSid', 'None')
            sAMAccountName = safe_ldap_attr(entry, 'sAMAccountName', 'None')
            accountExpires = safe_ldap_attr(entry, 'accountExpires', 'None')
            servicePrincipalName = safe_ldap_attr(entry, 'servicePrincipalName', 'None')
            badPwdCount = safe_ldap_attr(entry, 'badPwdCount', 'None')
            lastLogon = safe_ldap_attr(entry, 'lastLogon', 'None')
            lastLogoff = safe_ldap_attr(entry, 'lastLogoff', 'None')
            userAccountControl = fmt_uac(safe_ldap_attr(entry, 'userAccountControl', 'None')) or safe_ldap_attr(entry, 'userAccountControl', 'None')

            result = f"""Username: {sAMAccountName}
Description: {description}
Member of: {memberOf}
SID: {objectSid}
Account Expires: {accountExpires}
servicePrincipalName: {servicePrincipalName}
badPwdCount: {badPwdCount}
Last Logon: {lastLogon}
Last Logoff: {lastLogoff}
userAccountControl: {userAccountControl}"""
            
            values.append(result)
            console.print(result, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as f:
                for v in values:
                    f.write(v + "\n")
            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values



from impacket.dcerpc.v5 import transport, srvs
from impacket.dcerpc.v5.rpcrt import DCERPCException
from impacket.smbconnection import SMBConnection
from rich.console import Console
console = Console()

from handlers.profile import get_username, get_password, get_domain

def get_rpc_connection(host: str):
        """ RPC Connection Handler: establish an RPC connection with the host and returns DCE object """
        
        username, password, domain = get_username(), get_password(), get_domain()

        if len(password) == 32 and all(c in "0123456789abcdefABCDEF" for c in password):
            use_ntlmv2_hash = True
        else:
            use_ntlmv2_hash = False

        rpc_string = r"ncacn_np:%s[\pipe\srvsvc]" % host
        transport_obj = transport.DCERPCTransportFactory(rpc_string)

        if use_ntlmv2_hash:
            smb_connection = SMBConnection(host, host)
            lmhash = 'aad3b435b51404eeaad3b435b51404ee'
            nthash = password
            smb_connection.login(username, '', domain=domain, lmhash=lmhash, nthash=nthash)
            transport_obj.set_smb_connection(smb_connection)
        else:
            transport_obj.set_credentials(username, password, domain)

        try:
            transport_obj.connect()
            dce = transport_obj.DCERPC_class(transport_obj)
            dce.bind(srvs.MSRPC_UUID_SRVS)
            return dce 

        except DCERPCException as error:
            console.print(f"[[red]![/]] Error estabilishing RPC connection: {error}")
            return None


from rich.console import Console
console = Console()
from uuid import uuid4

from protocols.ldap import safe_ldap_attr

class Servers:
    name = "servers"
    desc = "Get 'sAMAccountName', 'operatingSystem' and 'dnsHostName' from all Windows Servers"
    search_filter = "(&(objectCategory=computer)(operatingSystem=*server*))"
    attributes = ["sAMAccountName", "operatingSystem", "dNSHostName"]

    def on_login(self, conn, base_dn, save_output = False, module_args = None):
        console.print(f"[[green]+[/]] [cyan]MODULE[/]  Running [yellow]{self.name}[/] module")

        results = conn.search(base_dn, self.search_filter, attributes=self.attributes)
        entries = conn.entries

        if not entries:
            console.print(f"\n[[red]![/]] No results for {self.name}") 
            return
        
        console.print(f"[[green]+[/]] [cyan]QUERY[/]   [white]{self.search_filter}[/]\n", highlight=False)

        values = []
        for entry in conn.entries:
            sAMAccountName = safe_ldap_attr(entry, 'sAMAccountName', 'None')
            operatingSystem = safe_ldap_attr(entry, 'operatingSystem', 'None')
            dNSHostName = safe_ldap_attr(entry, 'dNSHostName', 'None')

            result = f"{sAMAccountName} - {operatingSystem} - {dNSHostName}"
            values.append(result)

            console.print(result, highlight=False)

        if save_output:
            filename = f"{self.name}_{uuid4().hex}.txt"
            with open(filename, "w", encoding="utf-8") as file:
                for value in values:
                    file.write(value + "\n")

            console.print(f"\n[[green]+[/]] Output saved to {filename}", highlight=False)

        return values
